==> edge-exchange-plugins-0.11.12.tgz: package/lib/swap/totle.js <==
// 

import { div, mul } from 'biggystring'
import {









  InsufficientFundsError,
  NoAmountSpecifiedError,
  SwapCurrencyError
} from 'edge-core-js/types'

const pluginId = 'totle'
const swapInfo = {
  pluginId,
  displayName: 'Totle',

  orderUri: 'https://api.totle.com/swap',
  supportEmail: 'support@totle.com'
}

const swapUri = 'https://api.totle.com/swap'
const tokenUri = 'https://api.totle.com/tokens'
const expirationMs = 1000 * 60 * 20























































function checkReply(reply, request) {
  if (reply.success === false) {
    const code = reply.response.code
    // unsupported tokens
    if (code === 1203) {
      throw new SwapCurrencyError(
        swapInfo,
        request.fromCurrencyCode,
        request.toCurrencyCode
      )
    } else if (code === 3100) {
      throw new InsufficientFundsError()
    } else if (code === 1201) {
      throw new NoAmountSpecifiedError()
    }
  }
}

export function makeTotlePlugin(opts) {
  const { initOptions, io, log } = opts
  const { fetchCors = io.fetch } = io

  const { partnerContract, apiKey } = initOptions

  async function call(json) {
    const body = JSON.stringify({
      ...json,
      partnerContract,
      apiKey
    })

    log('call:', json)
    const headers = {
      'Content-Type': 'application/json'
    }
    const response = await fetchCors(swapUri, { method: 'POST', body, headers })
    if (!response.ok) {
      throw new Error(`Totle returned error code ${response.status}`)
    }
    const out = await response.json()
    log('swap reply:', out)
    return out
  }

  async function fetchTokens() {
    const response = await fetchCors(tokenUri, { method: 'GET' })
    if (!response.ok) {
      throw new Error(`Totle returned error code ${response.status}`)
    }
    const json = await response.json()
    const out = json.tokens
    log('token reply:', out)
    return out
  }

  const out = {
    swapInfo,

    async fetchSwapQuote(
      request,
      userSettings
    ) {
      const tokens = await fetchTokens()

      const fromToken = tokens.find(t => t.symbol === request.fromCurrencyCode)
      const toToken = tokens.find(t => t.symbol === request.toCurrencyCode)
      if (!fromToken || !toToken || fromToken.symbol === toToken.symbol) {
        throw new SwapCurrencyError(
          swapInfo,
          request.fromCurrencyCode,
          request.toCurrencyCode
        )
      }

      // Grab addresses:
      const {
        publicAddress: userFromAddress
      } = await request.fromWallet.getReceiveAddress({}) //  currencyCode ?
      const {
        publicAddress: userToAddress
      } = await request.toWallet.getReceiveAddress({}) //  currencyCode ?

      // Get the estimate from the server:
      const reply = await call({
        address: userFromAddress,
        config: {
          transactions: true
        },
        swap: {
          sourceAsset: fromToken.address,
          destinationAsset: toToken.address,
          [request.quoteFor === 'from'
            ? 'sourceAmount'
            : 'destinationAmount']: request.nativeAmount,
          // strictDestination: request.quoteFor === 'to',
          destinationAddress: userToAddress
        }
      })
      checkReply(reply, request)

      const { summary, transactions } = reply.response

      let fromNativeAmount = summary[0].sourceAmount // string with many zeroes
      let toNativeAmount = summary[0].destinationAmount // string with many zeroes
      const fromMultiplier = '1' + '0'.repeat(fromToken.decimals)
      const toMultiplier = '1' + '0'.repeat(toToken.decimals)
      const isSourceRequest = request.quoteFor === 'from'
      if (isSourceRequest) {
        const fromExchangeAmount = div(fromNativeAmount, fromMultiplier, 10)
        const toExchangeAmount = mul(
          fromExchangeAmount,
          summary[0].guaranteedRate
        )
        toNativeAmount = mul(toExchangeAmount, toMultiplier)
      } else {
        const toExchangeAmount = div(toNativeAmount, toMultiplier, 10)
        const fromExchangeAmount = mul(
          toExchangeAmount,
          summary[0].guaranteedRate
        )
        fromNativeAmount = mul(fromExchangeAmount, fromMultiplier)
      }

      const txs = []
      for (const tx of transactions) {
        // Make the transaction:
        const spendInfo = {
          currencyCode: request.fromCurrencyCode,
          spendTargets: [
            {
              nativeAmount: tx.tx.value,
              publicAddress: tx.tx.to,
              uniqueIdentifier: tx.id,
              otherParams: { data: tx.tx.data }
            }
          ],
          networkFeeOption: 'custom',
          customNetworkFee: {
            gasLimit: tx.tx.gas,
            gasPrice: String(parseInt(tx.tx.gasPrice) / 1000000000)
          },
          swapData: {
            isEstimate: false,
            payoutAddress: userToAddress,
            payoutCurrencyCode: request.toCurrencyCode,
            payoutNativeAmount: toNativeAmount,
            payoutWalletId: request.toWallet.id,
            plugin: { ...swapInfo }
          }
        }

        const transaction = await request.fromWallet.makeSpend(
          spendInfo
        )

        txs.push(transaction)
      }

      const quote = makeTotleSwapPluginQuote(
        request,
        fromNativeAmount,
        toNativeAmount,
        txs,
        userToAddress,
        false,
        new Date(Date.now() + expirationMs),
        log
      )
      log(quote)
      return quote
    }
  }

  return out
}

function makeTotleSwapPluginQuote(
  request,
  fromNativeAmount,
  toNativeAmount,
  txs,
  destinationAddress,
  isEstimate,
  expirationDate,
  log
) {
  log(arguments)
  const { fromWallet } = request
  const swapTx = txs[txs.length - 1]

  const out = {
    fromNativeAmount,
    toNativeAmount,
    networkFee: {
      currencyCode: fromWallet.currencyInfo.currencyCode,
      nativeAmount:
        swapTx.parentNetworkFee != null
          ? swapTx.parentNetworkFee
          : swapTx.networkFee
    },
    destinationAddress,
    pluginId,
    expirationDate,
    isEstimate,

    async approve() {
      let swapTx
      let index = 0
      for (const tx of txs) {
        const signedTransaction = await fromWallet.signTx(tx)
        // NOTE: The swap transaction will always be the last one
        swapTx = await fromWallet.broadcastTx(signedTransaction)
        const lastTransactionIndex = txs.length - 1
        // if it's the last transaction of the array then assign `nativeAmount` data
        // (after signing and broadcasting) for metadata purposes
        if (index === lastTransactionIndex) {
          tx.nativeAmount = `-${fromNativeAmount}`
        }
        await fromWallet.saveTx(signedTransaction)
        index++
      }
      if (!swapTx) throw new Error('No Totle swapTx')
      return {
        transaction: swapTx,
        destinationAddress,
        orderId: swapTx.txid
      }
    },

    async close() {}
  }
  return out
}

==> edge-exchange-plugins-0.11.36.tgz: package/lib/swap/totle.js <==
// 

import { div, mul } from 'biggystring'
import {
  asArray,
  asBoolean,
  asEither,
  asNumber,
  asObject,
  asOptional,
  asString,
  asValue
} from 'cleaners'
import {









  InsufficientFundsError,
  NoAmountSpecifiedError,
  SwapCurrencyError
} from 'edge-core-js/types'

const pluginId = 'totle'
const swapInfo = {
  pluginId,
  displayName: 'Totle',

  orderUri: 'https://api.totle.com/swap',
  supportEmail: 'support@totle.com'
}

const swapUri = 'https://api.totle.com/swap'
const tokenUri = 'https://api.totle.com/tokens'
const expirationMs = 1000 * 60 * 20

const asTotleErrorResponse = asObject({
  success: asValue(false),
  response: asObject({
    id: asString,
    code: asNumber,
    message: asString,
    info: asString,
    name: asOptional(asString),
    link: asOptional(asString)
  })
})

const asTotleSwapResponse = asObject({
  success: asValue(true),
  response: asObject({
    summary: asArray(
      asObject({
        destinationAmount: asString,
        guaranteedRate: asString,
        sourceAmount: asString
      })
    ),
    transactions: asArray(
      asObject({
        type: asEither(asValue('swap'), asValue('approve')),
        tx: asObject({
          to: asString,
          from: asString,
          value: asString,
          data: asString,
          gasPrice: asString,
          gas: asString,
          nonce: asOptional(asNumber)
        })
      })
    )
  })
})



// /swap
const asTotleSwapReply = asEither(asTotleSwapResponse, asTotleErrorResponse)


// /tokens
const asToken = asObject({
  name: asString,
  symbol: asString,
  decimals: asNumber,
  address: asString,
  tradable: asBoolean,
  iconUrl: asString
})


const asTotleTokensResponse = asObject({
  tokens: asArray(asToken)
})

function checkSwapReply(
  reply,
  request
) {
  // Handle error response
  if (reply.success === false) {
    const code = reply.response.code
    switch (code) {
      case 1203: // TokenNotFoundError
        throw new SwapCurrencyError(
          swapInfo,
          request.fromCurrencyCode,
          request.toCurrencyCode
        )
      case 3100: // InsufficientFundsError
        throw new InsufficientFundsError()
      case 1201: // AmountError
        throw new NoAmountSpecifiedError()
      default:
        throw new Error(`Totle API Error: ${reply.response.message}`)
    }
  }

  // Return swap response
  return reply
}

export function makeTotlePlugin(opts) {
  const { initOptions, io, log } = opts
  const { fetchCors = io.fetch } = io

  const { partnerContract, apiKey } = initOptions

  async function fetchSwapReply(json) {
    const body = JSON.stringify({
      ...json,
      partnerContract,
      apiKey
    })

    log('fetchSwap:', json)
    const headers = {
      'Content-Type': 'application/json'
    }
    const response = await fetchCors(swapUri, { method: 'POST', body, headers })
    if (!response.ok) {
      throw new Error(`Totle returned error code ${response.status}`)
    }
    const responseBody = await response.json()
    const reply = asTotleSwapReply(responseBody)
    log('swap reply:', reply)
    return reply
  }

  async function fetchTokens() {
    const response = await fetchCors(tokenUri, { method: 'GET' })
    if (!response.ok) {
      throw new Error(`Totle returned error code ${response.status}`)
    }
    const reply = asTotleTokensResponse(await response.json())
    const { tokens } = reply
    log('token reply:', tokens)
    return tokens
  }

  const out = {
    swapInfo,

    async fetchSwapQuote(
      request,
      userSettings
    ) {
      const tokens = await fetchTokens()

      const fromToken = tokens.find(t => t.symbol === request.fromCurrencyCode)
      const toToken = tokens.find(t => t.symbol === request.toCurrencyCode)
      if (!fromToken || !toToken || fromToken.symbol === toToken.symbol) {
        throw new SwapCurrencyError(
          swapInfo,
          request.fromCurrencyCode,
          request.toCurrencyCode
        )
      }

      // Grab addresses:
      const {
        publicAddress: userFromAddress
      } = await request.fromWallet.getReceiveAddress({}) //  currencyCode ?
      const {
        publicAddress: userToAddress
      } = await request.toWallet.getReceiveAddress({}) //  currencyCode ?

      // Get the estimate from the server:
      const reply = await fetchSwapReply({
        address: userFromAddress,
        config: {
          transactions: true
        },
        swap: {
          sourceAsset: fromToken.address,
          destinationAsset: toToken.address,
          [request.quoteFor === 'from'
            ? 'sourceAmount'
            : 'destinationAmount']: request.nativeAmount,
          // strictDestination: request.quoteFor === 'to',
          destinationAddress: userToAddress
        }
      })
      const swapResponse = checkSwapReply(reply, request)
      const { summary: summaries, transactions } = swapResponse.response
      const summary = summaries[0]

      let fromNativeAmount = summary.sourceAmount // string with many zeroes
      let toNativeAmount = summary.destinationAmount // string with many zeroes
      const fromMultiplier = '1' + '0'.repeat(fromToken.decimals)
      const toMultiplier = '1' + '0'.repeat(toToken.decimals)
      const isSourceRequest = request.quoteFor === 'from'
      if (isSourceRequest) {
        const fromExchangeAmount = div(fromNativeAmount, fromMultiplier, 10)
        const toExchangeAmount = mul(fromExchangeAmount, summary.guaranteedRate)
        toNativeAmount = mul(toExchangeAmount, toMultiplier)
      } else {
        const toExchangeAmount = div(toNativeAmount, toMultiplier, 10)
        const fromExchangeAmount = mul(toExchangeAmount, summary.guaranteedRate)
        fromNativeAmount = mul(fromExchangeAmount, fromMultiplier)
      }

      const txs = []
      for (const swapTansaction of transactions) {
        // Make the transaction:
        const spendInfo = {
          currencyCode: request.fromCurrencyCode,
          spendTargets: [
            {
              nativeAmount: swapTansaction.tx.value,
              publicAddress: swapTansaction.tx.to,
              uniqueIdentifier: swapTansaction.id,
              otherParams: { data: swapTansaction.tx.data }
            }
          ],
          networkFeeOption: 'custom',
          customNetworkFee: {
            gasLimit: swapTansaction.tx.gas,
            gasPrice: String(parseInt(swapTansaction.tx.gasPrice) / 1000000000)
          },
          swapData: {
            isEstimate: false,
            payoutAddress: userToAddress,
            payoutCurrencyCode: request.toCurrencyCode,
            payoutNativeAmount: toNativeAmount,
            payoutWalletId: request.toWallet.id,
            plugin: { ...swapInfo }
          }
        }

        const edgeTransaction = await request.fromWallet.makeSpend(
          spendInfo
        )

        txs.push(edgeTransaction)
      }

      const quote = makeTotleSwapPluginQuote(
        request,
        fromNativeAmount,
        toNativeAmount,
        txs,
        userToAddress,
        false,
        new Date(Date.now() + expirationMs),
        log
      )
      log(quote)
      return quote
    }
  }

  return out
}

function makeTotleSwapPluginQuote(
  request,
  fromNativeAmount,
  toNativeAmount,
  txs,
  destinationAddress,
  isEstimate,
  expirationDate,
  log
) {
  log(arguments)
  const { fromWallet } = request
  const swapTx = txs[txs.length - 1]

  const out = {
    fromNativeAmount,
    toNativeAmount,
    networkFee: {
      currencyCode: fromWallet.currencyInfo.currencyCode,
      nativeAmount:
        swapTx.parentNetworkFee != null
          ? swapTx.parentNetworkFee
          : swapTx.networkFee
    },
    destinationAddress,
    pluginId,
    expirationDate,
    isEstimate,

    async approve() {
      let swapTx
      let index = 0
      for (const tx of txs) {
        const signedTransaction = await fromWallet.signTx(tx)
        // NOTE: The swap transaction will always be the last one
        swapTx = await fromWallet.broadcastTx(signedTransaction)
        const lastTransactionIndex = txs.length - 1
        // if it's the last transaction of the array then assign `nativeAmount` data
        // (after signing and broadcasting) for metadata purposes
        if (index === lastTransactionIndex) {
          tx.nativeAmount = `-${fromNativeAmount}`
        }
        await fromWallet.saveTx(signedTransaction)
        index++
      }
      if (!swapTx) throw new Error('No Totle swapTx')
      return {
        transaction: swapTx,
        destinationAddress,
        orderId: swapTx.txid
      }
    },

    async close() {}
  }
  return out
}
